import contextlib
import os
import subprocess
import time

from distroinfo import exception


@contextlib.contextmanager
def cdir(path):
    prev_cwd = os.getcwd()
    os.chdir(path)
    try:
        yield
    finally:
        os.chdir(prev_cwd)


def ensure_dir(path):
    if os.path.exists(path):
        if not os.path.isdir(path):
            raise exception.NotADirectory(path=path)
    else:
        os.makedirs(path)


def get_default_cache_base_path():
    return os.path.expanduser(u"~/.distroinfo/cache")


def get_file_age(path):
    t_mod = os.path.getctime(path)
    t_now = time.time()
    return t_now - t_mod


def git(*cmd):
    cmd = ['git'] + list(cmd)
    try:
        prc = subprocess.Popen(cmd,
                               stdout=subprocess.PIPE,
                               stderr=subprocess.PIPE)
    except OSError:
        raise exception.CommandNotFound(cmd=cmd[0])
    out, err = prc.communicate()
    if prc.returncode != 0:
        raise exception.CommandFailed(cmd=" ".join(cmd),
                                      code=prc.returncode)
    return out.decode('utf-8')



import io
import os
import logging

from distroinfo import query


ASSETS_PATH = os.path.join(os.path.dirname(__file__), 'assets')
INFO_ASSETS_PATH = os.path.join(ASSETS_PATH, 'info')


def get_test_info_path(name):
    return os.path.join(INFO_ASSETS_PATH, name)


def capture_distroinfo_logger():
    log = logging.getLogger('distroinfo')
    log.setLevel(logging.DEBUG)
    log_stream = io.StringIO()
    h = logging.StreamHandler(log_stream)
    log.addHandler(h)
    return log_stream


def assert_dict_contains(tested, expected):
    for key, val in expected.items():
        assert key in tested
        assert tested[key] == val


def assert_rdoinfo_base(info):
    """
    Make sure data from rdoinfo rdo.yml file is parsed correctly
    by checking the flagship nova package.

    NOTE: this needs to be kept in sync with rdoinfo
    """
    pkg = query.get_package(info, 'openstack-nova')
    assert pkg
    nova = {
        'name': 'openstack-nova',
        'project': 'nova',
        'conf': 'rpmfactory-core',
        'upstream':
            'git://git.openstack.org/openstack/nova',
        'patches':
            'http://review.rdoproject.org/r/p/openstack/nova.git',
        'distgit':
            'https://github.com/rdo-packages/nova-distgit.git',
        'master-distgit':
            'https://github.com/rdo-packages/nova-distgit.git',
        'review-origin':
            'ssh://review.rdoproject.org:29418/openstack/nova-distgit.git',
        'review-patches':
            'ssh://review.rdoproject.org:29418/openstack/nova.git',
    }
    assert_dict_contains(pkg, expected=nova)
    assert 'tags' in pkg
    assert 'buildsys-tags' in pkg
    assert 'maintainers' in pkg


def assert_rdoinfo_deps(info):
    """
    Make sure data from rdoinfo rdo.yml file is parsed correctly
    by checking python-sphinx dependency packages.

    NOTE: this needs to be kept in sync with rdoinfo
    """
    pkg = query.get_package(info, 'python-sphinx')
    assert pkg
    nova = {
        'name': 'python-sphinx',
        'project': 'python-sphinx',
        'conf': 'rdo-dependency',
        'upstream': 'https://github.com/sphinx-doc/sphinx',
        'patches': None,
        'distgit': 'https://github.com/rdo-common/python-sphinx.git',
    }
    assert_dict_contains(pkg, expected=nova)
    assert 'tags' in pkg
    assert 'dependency' in pkg['tags']
    assert 'buildsys-tags' in pkg
    assert 'maintainers' in pkg


def assert_rdoinfo_full(info):
    assert_rdoinfo_base(info)
    assert_rdoinfo_deps(info)



import hashlib
import logging
import os
import requests
import yaml

from distroinfo import exception
from distroinfo import helpers
from distroinfo import repoman


logging.basicConfig(level=logging.ERROR)
log = logging.getLogger("distroinfo")


def get_id(s, postfix=False):
    h = hashlib.sha1(s.encode()).hexdigest()[:4]
    if postfix:
        h = '-' + h
    return h


class InfoFetcher(object):
    """
    Abstract class to derive simple info fetchers from.

    Implement get_file_content() to return info file contents.

    See CachedInfoFetcher if you also need caching.
    """
    def __init__(self, source, allow_import=True):
        self.source = source
        self.allow_import = allow_import
        self.fetching = set()
        # DistroInfo will set this to DistroInfo class in order to allow
        # recursive fetch without circular dependency.
        self.di_class = None

    def get_file_content(self, fn):
        raise NotImplementedError()

    def get_file_data(self, fn):
        content = self.get_file_content(fn)
        return yaml.load(content)

    def fetch(self, *info_files, **kwargs):
        contents = []
        remote_info = kwargs.get('remote_info', {})
        for ifn in info_files:
            if isinstance(ifn, dict):
                # fetch remote info using a new DistroInfo instance
                remote = next(iter(ifn))
                fn = ifn[remote]
                ri = remote_info.get(remote)
                if not ri:
                    raise exception.InvalidRemoteInfoRef(remote=remote)
                ri = ri.copy()
                ri['info_files'] = '__remote_info__'
                # this is the same as DistroInfo(**ri)
                remote_di = self.di_class(**ri)
                remote_contents = remote_di.fetcher.fetch(fn)
                contents += remote_contents
            else:
                # import within this info repo
                if ifn in self.fetching:
                    raise exception.CircularInfoInclude(info=ifn)
                info = self.get_file_data(ifn)
                contents.append(info)
                # collect remote-infos
                remotes = info.get('remote-info', {})
                if remotes:
                    remote_info.update(remotes)
                # handle includes recursively
                imports = info.get('import', [])
                if self.allow_import and imports:
                    self.fetching.add(ifn)
                    try:
                        contents += self.fetch(*imports,
                                               remote_info=remote_info)
                    finally:
                        self.fetching.remove(ifn)
        return contents


class CachedInfoFetcher(InfoFetcher):
    """
    Abstract class to derive caching info fetchers from.

    Implement get_file_content() to return info file contents.

    Only cache if self.cache_ttl > 0.
    """
    def __init__(self, *args, **kwargs):
        self.cache_ttl = kwargs.pop('cache_ttl', 0)
        self.cache_base_path = kwargs.pop('cache_base_path', None)
        if not self.cache_base_path:
            self.cache_base_path = helpers.get_default_cache_base_path()
        super(CachedInfoFetcher, self).__init__(*args, **kwargs)

    def get_file_content(self, fn):
        raise NotImplementedError()


class LocalInfoFetcher(InfoFetcher):
    """Fetch info files from local directory (source)"""
    def get_file_content(self, fn):
        fn = os.path.join(self.source, fn)
        return open(fn).read()


class RemoteInfoFetcher(CachedInfoFetcher):
    """
    Fetch remote info files from URL (source)

    Cache info files locally if self.cache_ttl > 0
    """
    def __init__(self, *args, **kwargs):
        super(RemoteInfoFetcher, self).__init__(*args, **kwargs)
        self.cache_path = os.path.join(
            self.cache_base_path, get_id(self.source))

    def fetch_file(self, fn):
        url = u'%s%s' % (self.source, fn)
        log.info(u'Fetching remote file: %s' % url)
        req = requests.get(url)
        if req.ok:
            if self.cache_ttl:
                # cache this file
                helpers.ensure_dir(self.cache_path)
                path = os.path.join(self.cache_path, fn)
                open(path, 'wt').write(req.text)
            return req.text
        else:
            raise exception.RemoteFetchError(
                code=req.status_code, reason=req.reason, url=url)

    def get_file_content(self, fn):
        path = os.path.join(self.cache_path, fn)
        fetch = True
        if self.cache_ttl and os.path.exists(path):
            # look for cache first
            age = helpers.get_file_age(path)
            if age <= self.cache_ttl:
                # use cached version
                fetch = False
                log.info(u'Using %d s old cached version of %s' % (age, fn))
        if fetch:
            text = self.fetch_file(fn)
        else:
            text = open(path, 'rt').read()
        return text


class RemoteGitInfoFetcher(CachedInfoFetcher):
    """
    Fetch info files from a remote git repo (source)

    Use git clone to get the repository.

    Only sync the repo if local copy is older than self.cache_ttl
    """
    def __init__(self, *args, **kwargs):
        super(RemoteGitInfoFetcher, self).__init__(*args, **kwargs)
        self.repo = repoman.GitRepoManager(
            url=self.source,
            ttl=self.cache_ttl,
            base_path=self.cache_base_path,
            repo_dir_postfix=get_id(self.source, postfix=True))
        self.synced = False

    def get_file_content(self, fn):
        if not self.synced:
            self.repo.sync()
            self.synced = True
        path = self.repo.get_file_path(fn)
        return open(path).read()

    @property
    def cache_path(self):
        return self.repo.repo_path



from distroinfo.info import DistroInfo
from distroinfo.query import get_package
from distroinfo.query import get_release

import tests.test_common as common


def test_remote_info():
    di = DistroInfo('remote.yml',
                    local_info=common.get_test_info_path('minimal'))
    info = di.get_info()

    edf = info['package-default'].get('extra-default-field')
    assert edf == 'foo-default-%(project)s'

    nova = get_package(info, 'openstack-nova')
    assert nova['extra-default-field'] == 'foo-default-nova'
    assert nova['extra-conf-field'] == 'foo-conf-nova'
    assert nova['master-distgit'] == 'overriden-nova'
    assert 'extra@maintain.er' in nova['maintainers']

    rocky = get_release(info, 'rocky')
    assert rocky['branch'] == 'overriden-branch'



import logging
import os
import re
import shutil
import time

from distroinfo import exception
from distroinfo import helpers
from distroinfo.helpers import git


logging.basicConfig(level=logging.ERROR)
log = logging.getLogger("distroinfo")


class GitRepoManager(object):
    repo_desc = u'git'

    def __init__(self,
                 base_path,
                 url,
                 repo_dir_postfix=None,
                 ttl=None):
        self.base_path = os.path.abspath(base_path)
        self.url = url
        self.repo_dir_postfix = repo_dir_postfix
        _, _, self.repo_name = url.rpartition('/')
        self.ttl = ttl
        if not self.repo_name:
            raise exception.RepoError(
                what=u"Failed to parse %s repo URL: %s" % (self.repo_desc,
                                                           self.url))
        self.repo_path = os.path.join(
            self.base_path,
            self.repo_name + self.repo_dir_postfix or '')

    def nuke(self):
        log.warn(u"Removing %s repo: %s" % (self.repo_desc, self.repo_path))
        shutil.rmtree(self.repo_path, ignore_errors=True)

    def clone(self):
        log.info(u"Cloning {desc} repo: {url}\n"
                 u"        {space} into: {path}".format(
                     desc=self.repo_desc,
                     space=len(self.repo_desc) * ' ',
                     url=self.url,
                     path=self.repo_path))
        git('clone', self.url, self.repo_path)

    def get_last_fetch_time(self):
        path = os.path.join(self.repo_path, '.git/FETCH_HEAD')
        if os.path.exists(path):
            return os.path.getmtime(path)
        # FETCH_HEAD isn't available on fresh clone
        path = os.path.join(self.repo_path, '.git')
        if os.path.exists(path):
            return os.path.getmtime(path)
        return None

    def fetch(self, force=False):
        need_fetch = True
        with self.repo_dir():
            delta = 0
            if not force and self.ttl:
                try:
                    # caching enabled, check for last repo fetch
                    t_fetch = self.get_last_fetch_time()
                    delta = int(time.time()) - t_fetch
                    if delta < self.ttl:
                        need_fetch = False
                        log.info(u"Existing %s repo is fresh enough, "
                                 u"it was fetched %d s ago: %s" % (
                                     self.repo_desc, delta, self.repo_path))
                    else:
                        log.info(u"Existing %s repo is too old, "
                                 u"it was fetched %d s ago: %s" % (
                                     self.repo_desc, delta, self.repo_path))
                except Exception as e:
                    raise e
            if need_fetch:
                log.info(u"Fetching %s repo: %s" % (self.repo_desc,
                                                    self.repo_path))
                git('fetch', 'origin')
                git('checkout', '-f', 'master')
                git('reset', '--hard', 'origin/master')

    def repo_dir(self):
        """
        Use

            with self.repo_dir():
                commands

        to temporarily set current directory to repo path"""
        return helpers.cdir(self.repo_path)

    def get_file_path(self, fn):
        return os.path.join(self.repo_path, fn)

    def check_remote(self):
        assert self.url
        with self.repo_dir():
            remotes = git('remote', '-v')
        pattern = r'^origin\s+%s\s+\(fetch\)$' % re.escape(self.url)
        if not re.search(pattern, remotes, re.MULTILINE):
            raise exception.RepoError(what="origin isn't set to expected URL: "
                                           "%s" % self.url)

    def sync(self, force_fetch=False):
        if not self.url:
            if not os.path.isdir(self.repo_path):
                raise exception.NotADirectory(path=self.repo_path)
            return
        if self.base_path and not os.path.isdir(self.base_path):
            log.info(u"Creating base directory: %s" % self.base_path)
            os.makedirs(self.base_path)
        if not os.path.isdir(self.repo_path):
            self.clone()
        else:
            try:
                self.check_remote()
            except exception.RepoError as e:
                log.warn(u"%s repo didn't pass the checks, renewing: %s"
                         % (self.repo_desc, e))
                self.nuke()
                self.clone()
            else:
                self.fetch(force=force_fetch)



from distroinfo.info import DistroInfo
from distroinfo import parse

import tests.test_common as common


def test_rdoinfo_base():
    di = DistroInfo('rdo.yml',
                    local_info=common.get_test_info_path('rdoinfo'))
    info = di.get_info()
    common.assert_rdoinfo_base(info)


def test_rdoinfo_full():
    di = DistroInfo('rdo-full.yml',
                    local_info=common.get_test_info_path('rdoinfo'))
    info = di.get_info()
    common.assert_rdoinfo_full(info)


def test_rdoinfo_merge():
    di = DistroInfo(['rdo.yml', 'deps.yml'],
                    local_info=common.get_test_info_path('rdoinfo'))
    info = di.get_info()
    common.assert_rdoinfo_full(info)


def test_dict_conversion():
    di = DistroInfo('rdo.yml',
                    local_info=common.get_test_info_path('rdoinfo'))
    info = di.get_info()
    info_dicts = di.get_info(info_dicts=True)
    assert parse.info2dicts(info) == info_dicts
    assert parse.info2lists(info_dicts) == info
